import os
import pytest
import tornado

from .utils import expected_http_error

from nbformat import writes
from nbformat.v4 import (new_notebook,
                         new_markdown_cell, new_code_cell,
                         new_output)


async def test_hidden_files(fetch, serverapp, root_dir):
    not_hidden = [
        u'å b',
        u'å b/ç. d',
    ]
    hidden = [
        u'.å b',
        u'å b/.ç d',
    ]
    dirs = not_hidden + hidden

    for d in dirs:
        path  = root_dir / d.replace('/', os.sep)
        path.mkdir(parents=True, exist_ok=True)
        path.joinpath('foo').write_text('foo')
        path.joinpath('.foo').write_text('.foo')


    for d in not_hidden:
        path = root_dir / d.replace('/', os.sep)

        r = await fetch(
            'files', d, 'foo',
            method='GET'
        )
        assert r.body.decode() == 'foo'

        with pytest.raises(tornado.httpclient.HTTPClientError) as e:
            r = await fetch(
                'files', d, '.foo',
                method='GET'
            )
        assert expected_http_error(e, 404)


    for d in hidden:
        path = root_dir / d.replace('/', os.sep)
        for foo in ('foo', '.foo'):
            with pytest.raises(tornado.httpclient.HTTPClientError) as e:
                r = await fetch(
                    'files', d, foo,
                    method='GET'
                )
            assert expected_http_error(e, 404)

    serverapp.contents_manager.allow_hidden = True

    for d in not_hidden:
        path = root_dir / d.replace('/', os.sep)

        r = await fetch(
            'files', d, 'foo',
            method='GET'
        )
        assert r.body.decode() == 'foo'

        r = await fetch(
            'files', d, '.foo',
            method='GET'
        )
        assert r.body.decode() == '.foo'

        for d in hidden:
            path = root_dir / d.replace('/', os.sep)

            for foo in ('foo', '.foo'):
                r = await fetch(
                    'files', d, foo,
                    method='GET'
                )
                assert r.body.decode() == foo


async def test_contents_manager(fetch, serverapp, root_dir):
    "make sure ContentsManager returns right files (ipynb, bin, txt)."
    nb = new_notebook(
        cells=[
            new_markdown_cell(u'Created by test ³'),
            new_code_cell("print(2*6)", outputs=[
                new_output("stream", text="12"),
            ])
        ]
    )
    root_dir.joinpath('testnb.ipynb').write_text(writes(nb, version=4), encoding='utf-8')
    root_dir.joinpath('test.bin').write_bytes(b'\xff' + os.urandom(5))
    root_dir.joinpath('test.txt').write_text('foobar')

    r = await fetch(
        'files/testnb.ipynb',
        method='GET'
    )
    assert r.code == 200
    assert 'print(2*6)' in r.body.decode('utf-8')

    r = await fetch(
        'files/test.bin',
        method='GET'
    )
    assert r.code == 200
    assert r.headers['content-type'] == 'application/octet-stream'
    assert r.body[:1] == b'\xff'
    assert len(r.body) == 6

    r = await fetch(
        'files/test.txt',
        method='GET'
    )
    assert r.code == 200
    assert r.headers['content-type'] == 'text/plain; charset=UTF-8'
    assert r.body.decode() == 'foobar'


async def test_download(fetch, serverapp, root_dir):
    text = 'hello'
    root_dir.joinpath('test.txt').write_text(text)

    r = await fetch(
        'files', 'test.txt',
        method='GET'
    )
    disposition = r.headers.get('Content-Disposition', '')
    assert 'attachment' not in disposition

    r = await fetch(
        'files', 'test.txt',
        method='GET',
        params={'download': True}
    )
    disposition = r.headers.get('Content-Disposition', '')
    assert 'attachment' in disposition
    assert "filename*=utf-8''test.txt" in disposition


async def test_old_files_redirect(fetch, serverapp, root_dir):
    """pre-2.0 'files/' prefixed links are properly redirected"""
    root_dir.joinpath('files').mkdir(parents=True, exist_ok=True)
    root_dir.joinpath('sub', 'files').mkdir(parents=True, exist_ok=True)


    for prefix in ('', 'sub'):
        root_dir.joinpath(prefix, 'files', 'f1.txt').write_text(prefix + '/files/f1')
        root_dir.joinpath(prefix, 'files', 'f2.txt').write_text(prefix + '/files/f2')
        root_dir.joinpath(prefix, 'f2.txt').write_text(prefix + '/f2')
        root_dir.joinpath(prefix, 'f3.txt').write_text(prefix + '/f3')

    # These depend on the tree handlers
    #
    # def test_download(self):
    #     rootdir = self.root_dir

    #     text = 'hello'
    #     with open(pjoin(rootdir, 'test.txt'), 'w') as f:
    #         f.write(text)

    #     r = self.request('GET', 'files/test.txt')
    #     disposition = r.headers.get('Content-Disposition', '')
    #     self.assertNotIn('attachment', disposition)

    #     r = self.request('GET', 'files/test.txt?download=1')
    #     disposition = r.headers.get('Content-Disposition', '')
    #     self.assertIn('attachment', disposition)
    #     self.assertIn("filename*=utf-8''test.txt", disposition)



# Only Run tests on MacOS and Linux
import shutil
import pytest
import json
import asyncio
import sys

# Skip this whole module on Windows. The terminal API leads
# to timeouts on Windows CI.
if sys.platform.startswith('win'):
    pytest.skip("Terminal API tests time out on Windows.", allow_module_level=True)


# Kill all running terminals after each test to avoid cross-test issues
# with still running terminals.
@pytest.fixture
def kill_all(serverapp):
    async def _():
        await serverapp.web_app.settings["terminal_manager"].kill_all()
    return _


@pytest.fixture
def terminal_path(tmp_path):
    subdir = tmp_path.joinpath('terminal_path')
    subdir.mkdir()

    yield subdir

    shutil.rmtree(str(subdir), ignore_errors=True)


async def test_terminal_create(fetch, kill_all):
    await fetch(
        'api', 'terminals',
        method='POST',
        allow_nonstandard_methods=True,
    )

    resp_list = await fetch(
        'api', 'terminals',
        method='GET',
        allow_nonstandard_methods=True,
    )

    data = json.loads(resp_list.body.decode())

    assert len(data) == 1
    await kill_all()


async def test_terminal_create_with_kwargs(fetch, ws_fetch, terminal_path, kill_all):
    resp_create = await fetch(
        'api', 'terminals',
        method='POST',
        body=json.dumps({'cwd': str(terminal_path)}),
        allow_nonstandard_methods=True,
    )

    data = json.loads(resp_create.body.decode())
    term_name = data['name']

    resp_get = await fetch(
        'api', 'terminals', term_name,
        method='GET',
        allow_nonstandard_methods=True,
    )

    data = json.loads(resp_get.body.decode())

    assert data['name'] == term_name
    await kill_all()


async def test_terminal_create_with_cwd(fetch, ws_fetch, terminal_path):
    resp = await fetch(
        'api', 'terminals',
        method='POST',
        body=json.dumps({'cwd': str(terminal_path)}),
        allow_nonstandard_methods=True,
    )

    data = json.loads(resp.body.decode())
    term_name = data['name']

    ws = await ws_fetch(
        'terminals', 'websocket', term_name
    )

    ws.write_message(json.dumps(['stdin', 'pwd\r\n']))

    message_stdout = ''
    while True:
        try:
            message = await asyncio.wait_for(ws.read_message(), timeout=1.0)
        except asyncio.TimeoutError:
            break

        message = json.loads(message)

        if message[0] == 'stdout':
            message_stdout += message[1]

    ws.close()

    assert str(terminal_path) in message_stdout



import os
import sys
import json
import shutil
import pytest
import asyncio
from binascii import hexlify

import urllib.parse
import tornado
from tornado.escape import url_escape

from traitlets.config import Config

import jupyter_core.paths
from jupyter_server.extension import serverextension
from jupyter_server.serverapp import ServerApp
from jupyter_server.utils import url_path_join
from jupyter_server.services.contents.filemanager import FileContentsManager

import nbformat

# This shouldn't be needed anymore, since pytest_tornasync is found in entrypoints
pytest_plugins = "pytest_tornasync"

# NOTE: This is a temporary fix for Windows 3.8
# We have to override the io_loop fixture with an
# asyncio patch. This will probably be removed in
# the future.

@pytest.fixture
def asyncio_patch():
    ServerApp()._init_asyncio_patch()

@pytest.fixture
def io_loop(asyncio_patch):
    loop = tornado.ioloop.IOLoop()
    loop.make_current()
    yield loop
    loop.clear_current()
    loop.close(all_fds=True)


def mkdir(tmp_path, *parts):
    path = tmp_path.joinpath(*parts)
    if not path.exists():
        path.mkdir(parents=True)
    return path


server_config = pytest.fixture(lambda: {})
home_dir = pytest.fixture(lambda tmp_path: mkdir(tmp_path, "home"))
data_dir = pytest.fixture(lambda tmp_path: mkdir(tmp_path, "data"))
config_dir = pytest.fixture(lambda tmp_path: mkdir(tmp_path, "config"))
runtime_dir = pytest.fixture(lambda tmp_path: mkdir(tmp_path, "runtime"))
root_dir = pytest.fixture(lambda tmp_path: mkdir(tmp_path, "root_dir"))
template_dir = pytest.fixture(lambda tmp_path: mkdir(tmp_path, "templates"))
system_jupyter_path = pytest.fixture(
    lambda tmp_path: mkdir(tmp_path, "share", "jupyter")
)
env_jupyter_path = pytest.fixture(
    lambda tmp_path: mkdir(tmp_path, "env", "share", "jupyter")
)
system_config_path = pytest.fixture(lambda tmp_path: mkdir(tmp_path, "etc", "jupyter"))
env_config_path = pytest.fixture(
    lambda tmp_path: mkdir(tmp_path, "env", "etc", "jupyter")
)
some_resource = u"The very model of a modern major general"
sample_kernel_json = {
    'argv':['cat', '{connection_file}'],
    'display_name': 'Test kernel',
}
argv = pytest.fixture(lambda: [])

@pytest.fixture
def environ(
    monkeypatch,
    tmp_path,
    home_dir,
    data_dir,
    config_dir,
    runtime_dir,
    root_dir,
    system_jupyter_path,
    system_config_path,
    env_jupyter_path,
    env_config_path,
):
    monkeypatch.setenv("HOME", str(home_dir))
    monkeypatch.setenv("PYTHONPATH", os.pathsep.join(sys.path))

    # Get path to nbconvert template directory *before*
    # monkeypatching the paths env variable.
    possible_paths = jupyter_core.paths.jupyter_path('nbconvert', 'templates')
    nbconvert_path = None
    for path in possible_paths:
        if os.path.exists(path):
            nbconvert_path = path
            break

    nbconvert_target = data_dir / 'nbconvert' / 'templates'

    # monkeypatch.setenv("JUPYTER_NO_CONFIG", "1")
    monkeypatch.setenv("JUPYTER_CONFIG_DIR", str(config_dir))
    monkeypatch.setenv("JUPYTER_DATA_DIR", str(data_dir))
    monkeypatch.setenv("JUPYTER_RUNTIME_DIR", str(runtime_dir))
    monkeypatch.setattr(
        jupyter_core.paths, "SYSTEM_JUPYTER_PATH", [str(system_jupyter_path)]
    )
    monkeypatch.setattr(jupyter_core.paths, "ENV_JUPYTER_PATH", [str(env_jupyter_path)])
    monkeypatch.setattr(
        jupyter_core.paths, "SYSTEM_CONFIG_PATH", [str(system_config_path)]
    )
    monkeypatch.setattr(jupyter_core.paths, "ENV_CONFIG_PATH", [str(env_config_path)])

    # copy nbconvert templates to new tmp data_dir.
    if nbconvert_path:
        shutil.copytree(nbconvert_path, str(nbconvert_target))


@pytest.fixture
def extension_environ(env_config_path, monkeypatch):
    """Monkeypatch a Jupyter Extension's config path into each test's environment variable"""
    monkeypatch.setattr(serverextension, "ENV_CONFIG_PATH", [str(env_config_path)])


@pytest.fixture(scope='function')
def configurable_serverapp(
    environ,
    server_config,
    argv,
    http_port,
    tmp_path,
    root_dir,
    io_loop,
):
    ServerApp.clear_instance()

    def _configurable_serverapp(
        config=server_config,
        argv=argv,
        environ=environ,
        http_port=http_port,
        tmp_path=tmp_path,
        root_dir=root_dir,
        **kwargs
    ):
        c = Config(config)
        c.NotebookNotary.db_file = ":memory:"
        token = hexlify(os.urandom(4)).decode("ascii")
        url_prefix = "/"
        app = ServerApp.instance(
            # Set the log level to debug for testing purposes
            log_level='DEBUG',
            port=http_port,
            port_retries=0,
            open_browser=False,
            root_dir=str(root_dir),
            base_url=url_prefix,
            config=c,
            allow_root=True,
            token=token,
            **kwargs
        )

        app.init_signal = lambda: None
        app.log.propagate = True
        app.log.handlers = []
        # Initialize app without httpserver
        app.initialize(argv=argv, new_httpserver=False)
        app.log.propagate = True
        app.log.handlers = []
        # Start app without ioloop
        app.start_app()
        return app

    return _configurable_serverapp


@pytest.fixture(scope="function")
def serverapp(server_config, argv, configurable_serverapp):
    app = configurable_serverapp(config=server_config, argv=argv)
    yield app
    app.remove_server_info_file()
    app.remove_browser_open_file()
    app.cleanup_kernels()


@pytest.fixture
def app(serverapp):
    """app fixture is needed by pytest_tornasync plugin"""
    return serverapp.web_app


@pytest.fixture
def auth_header(serverapp):
    return {"Authorization": "token {token}".format(token=serverapp.token)}


@pytest.fixture
def http_port(http_server_port):
    return http_server_port[-1]


@pytest.fixture
def base_url():
    return "/"


@pytest.fixture
def fetch(http_server_client, auth_header, base_url):
    """fetch fixture that handles auth, base_url, and path"""
    def client_fetch(*parts, headers={}, params={}, **kwargs):
        # Handle URL strings
        path_url = url_escape(url_path_join(base_url, *parts), plus=False)
        params_url = urllib.parse.urlencode(params)
        url = path_url + "?" + params_url
        # Add auth keys to header
        headers.update(auth_header)
        # Make request.
        return http_server_client.fetch(
            url, headers=headers, request_timeout=20, **kwargs
        )
    return client_fetch


@pytest.fixture
def ws_fetch(auth_header, http_port):
    """websocket fetch fixture that handles auth, base_url, and path"""
    def client_fetch(*parts, headers={}, params={}, **kwargs):
        # Handle URL strings
        path = url_escape(url_path_join(*parts), plus=False)
        urlparts = urllib.parse.urlparse('ws://localhost:{}'.format(http_port))
        urlparts = urlparts._replace(
            path=path,
            query=urllib.parse.urlencode(params)
        )
        url = urlparts.geturl()
        # Add auth keys to header
        headers.update(auth_header)
        # Make request.
        req = tornado.httpclient.HTTPRequest(
            url,
            headers=auth_header,
            connect_timeout=120
        )
        return tornado.websocket.websocket_connect(req)
    return client_fetch


@pytest.fixture
def kernelspecs(data_dir):
    spec_names = ['sample', 'sample 2']
    for name in spec_names:
        sample_kernel_dir = data_dir.joinpath('kernels', name)
        sample_kernel_dir.mkdir(parents=True)
        # Create kernel json file
        sample_kernel_file = sample_kernel_dir.joinpath('kernel.json')
        sample_kernel_file.write_text(json.dumps(sample_kernel_json))
        # Create resources text
        sample_kernel_resources = sample_kernel_dir.joinpath('resource.txt')
        sample_kernel_resources.write_text(some_resource)


@pytest.fixture(params=[True, False])
def contents_manager(request, tmp_path):
    return FileContentsManager(root_dir=str(tmp_path), use_atomic_writing=request.param)


@pytest.fixture
def create_notebook(root_dir):
    """Create a notebook in the test's home directory."""
    def inner(nbpath):
        nbpath = root_dir.joinpath(nbpath)
        # Check that the notebook has the correct file extension.
        if nbpath.suffix != '.ipynb':
            raise Exception("File extension for notebook must be .ipynb")
        # If the notebook path has a parent directory, make sure it's created.
        parent = nbpath.parent
        parent.mkdir(parents=True, exist_ok=True)
        # Create a notebook string and write to file.
        nb = nbformat.v4.new_notebook()
        nbtext = nbformat.writes(nb, version=4)
        nbpath.write_text(nbtext)
    return inner



import pytest
import json

import tornado

from jupyter_server.pytest_plugin import some_resource

from jupyter_client.kernelspec import NATIVE_KERNEL_NAME

from ...utils import expected_http_error


async def test_list_kernelspecs_bad(fetch, kernelspecs, data_dir):
    bad_kernel_dir = data_dir.joinpath(data_dir, 'kernels', 'bad')
    bad_kernel_dir.mkdir(parents=True)
    bad_kernel_json = bad_kernel_dir.joinpath('kernel.json')
    bad_kernel_json.write_text('garbage')

    r = await fetch(
        'api', 'kernelspecs',
        method='GET'
    )
    model = json.loads(r.body.decode())
    assert isinstance(model, dict)
    assert model['default'] == NATIVE_KERNEL_NAME
    specs = model['kernelspecs']
    assert isinstance(specs, dict)
    assert len(specs) > 2


async def test_list_kernelspecs(fetch, kernelspecs):
    r = await fetch(
        'api', 'kernelspecs',
        method='GET'
    )
    model = json.loads(r.body.decode())
    assert isinstance(model, dict)
    assert model['default'] == NATIVE_KERNEL_NAME
    specs = model['kernelspecs']
    assert isinstance(specs, dict)
    assert len(specs) > 2

    def is_sample_kernelspec(s):
        return s['name'] == 'sample' and s['spec']['display_name'] == 'Test kernel'

    def is_default_kernelspec(s):
        return s['name'] == NATIVE_KERNEL_NAME and s['spec']['display_name'].startswith("Python")

    assert any(is_sample_kernelspec(s) for s in specs.values()), specs
    assert any(is_default_kernelspec(s) for s in specs.values()), specs


async def test_get_kernelspecs(fetch, kernelspecs):
    r = await fetch(
        'api', 'kernelspecs', 'Sample',
        method='GET'
    )
    model = json.loads(r.body.decode())
    assert model['name'].lower() == 'sample'
    assert isinstance(model['spec'], dict)
    assert model['spec']['display_name'] == 'Test kernel'
    assert isinstance(model['resources'], dict)


async def test_get_kernelspec_spaces(fetch, kernelspecs):
    r = await fetch(
        'api', 'kernelspecs', 'sample%202',
        method='GET'
    )
    model = json.loads(r.body.decode())
    assert model['name'].lower() == 'sample 2'


async def test_get_nonexistant_kernelspec(fetch, kernelspecs):
    with pytest.raises(tornado.httpclient.HTTPClientError) as e:
        await fetch(
            'api', 'kernelspecs', 'nonexistant',
            method='GET'
        )
    assert expected_http_error(e, 404)


async def test_get_kernel_resource_file(fetch, kernelspecs):
    r = await fetch(
        'kernelspecs', 'sAmple', 'resource.txt',
        method='GET'
    )
    res = r.body.decode('utf-8')
    assert res == some_resource


async def test_get_nonexistant_resource(fetch, kernelspecs):
    with pytest.raises(tornado.httpclient.HTTPClientError) as e:
        await fetch(
            'kernelspecs', 'nonexistant', 'resource.txt',
            method='GET'
        )
    assert expected_http_error(e, 404)

    with pytest.raises(tornado.httpclient.HTTPClientError) as e:
        await fetch(
            'kernelspecs', 'sample', 'nonexistant.txt',
            method='GET'
        )
    assert expected_http_error(e, 404)


import os
import pytest


# All test coroutines will be treated as marked.
pytestmark = pytest.mark.script_launch_mode('subprocess')


def test_server_extension_list(environ, script_runner):
    ret = script_runner.run(
        'jupyter',
        'server',
        'extension',
        'list',
    )
    assert ret.success


import pytest


@pytest.fixture
def server_config(template_dir):
    return {
            "ServerApp": {
                "jpserver_extensions": {
                    "tests.extension.mockextensions": True
                }
            },
            "MockExtensionApp": {
                "template_paths": [
                    str(template_dir)
                ]
            }
        }


async def test_handler(fetch):
    r = await fetch(
        'mock',
        method='GET'
    )
    assert r.code == 200
    assert r.body.decode() == 'mock trait'


async def test_handler_template(fetch, mock_template):
    r = await fetch(
        'mock_template',
        method='GET'
    )
    assert r.code == 200


@pytest.mark.parametrize(
    'server_config',
    [
        {
            "ServerApp": {
                "jpserver_extensions": {
                    "tests.extension.mockextensions": True
                }
            },
            "MockExtensionApp": {
                # Change a trait in the MockExtensionApp using
                # the following config value.
                "mock_trait": "test mock trait"
            }
        }
    ]
)
async def test_handler_setting(fetch):
    # Test that the extension trait was picked up by the webapp.
    r = await fetch(
        'mock',
        method='GET'
    )
    assert r.code == 200
    assert r.body.decode() == 'test mock trait'


@pytest.mark.parametrize(
    'argv', (['--MockExtensionApp.mock_trait="test mock trait"'],)
)
async def test_handler_argv(fetch):
    # Test that the extension trait was picked up by the webapp.
    r = await fetch(
        'mock',
        method='GET'
    )
    assert r.code == 200
    assert r.body.decode() == 'test mock trait'


@pytest.mark.parametrize(
    'server_config',
    [
        {
            "ServerApp": {
                "jpserver_extensions": {
                    "tests.extension.mockextensions": True
                },
                # Move extension handlers behind a url prefix
                "base_url": "test_prefix"
            },
            "MockExtensionApp": {
                # Change a trait in the MockExtensionApp using
                # the following config value.
                "mock_trait": "test mock trait"
            }
        }
    ]
)
async def test_base_url(fetch):
    # Test that the extension's handlers were properly prefixed
    r = await fetch(
        'test_prefix', 'mock',
        method='GET'
    )
    assert r.code == 200
    assert r.body.decode() == 'test mock trait'

    # Test that the static namespace was prefixed by base_url
    r = await fetch(
        'test_prefix',
        'static', 'mockextension', 'mock.txt',
        method='GET'
    )
    assert r.code == 200
    body = r.body.decode()
    assert "mock static content" in body



import os
import sys
import time
import pytest
import functools
from traitlets import TraitError
from tornado.web import HTTPError
from itertools import combinations


from nbformat import v4 as nbformat

from jupyter_server.services.contents.filemanager import FileContentsManager
from ...utils import expected_http_error

# -------------- Functions ----------------------------

def _make_dir(contents_manager, api_path):
    """
    Make a directory.
    """
    os_path = contents_manager._get_os_path(api_path)
    try:
        os.makedirs(os_path)
    except OSError:
        print("Directory already exists: %r" % os_path)


def symlink(contents_manager, src, dst):
    """Make a symlink to src from dst

    src and dst are api_paths
    """
    src_os_path = contents_manager._get_os_path(src)
    dst_os_path = contents_manager._get_os_path(dst)
    print(src_os_path, dst_os_path, os.path.isfile(src_os_path))
    os.symlink(src_os_path, dst_os_path)


def add_code_cell(notebook):
    output = nbformat.new_output("display_data", {'application/javascript': "alert('hi');"})
    cell = nbformat.new_code_cell("print('hi')", outputs=[output])
    notebook.cells.append(cell)


def new_notebook(contents_manager):
    cm = contents_manager
    model = cm.new_untitled(type='notebook')
    name = model['name']
    path = model['path']

    full_model = cm.get(path)
    nb = full_model['content']
    nb['metadata']['counter'] = int(1e6 * time.time())
    add_code_cell(nb)

    cm.save(full_model, path)
    return nb, name, path


def make_populated_dir(contents_manager, api_path):
    cm = contents_manager
    _make_dir(cm, api_path)
    cm.new(path="/".join([api_path, "nb.ipynb"]))
    cm.new(path="/".join([api_path, "file.txt"]))


def check_populated_dir_files(contents_manager, api_path):
    dir_model = contents_manager.get(api_path)

    assert dir_model['path'] == api_path
    assert dir_model['type'] == "directory"

    for entry in dir_model['content']:
        if entry['type'] == "directory":
            continue
        elif entry['type'] == "file":
            assert entry['name'] == "file.txt"
            complete_path = "/".join([api_path, "file.txt"])
            assert entry["path"] == complete_path
        elif entry['type'] == "notebook":
            assert entry['name'] == "nb.ipynb"
            complete_path = "/".join([api_path, "nb.ipynb"])
            assert entry["path"] == complete_path

# ----------------- Tests ----------------------------------

def test_root_dir(tmp_path):
    fm = FileContentsManager(root_dir=str(tmp_path))
    assert fm.root_dir == str(tmp_path)


def test_missing_root_dir(tmp_path):
    root = tmp_path / 'notebook' / 'dir' / 'is' / 'missing'
    with pytest.raises(TraitError):
        FileContentsManager(root_dir=str(root))


def test_invalid_root_dir(tmp_path):
    temp_file = tmp_path / 'file.txt'
    temp_file.write_text('')
    with pytest.raises(TraitError):
        FileContentsManager(root_dir=str(temp_file))

def test_get_os_path(tmp_path):
    fm = FileContentsManager(root_dir=str(tmp_path))
    path = fm._get_os_path('/path/to/notebook/test.ipynb')
    rel_path_list =  '/path/to/notebook/test.ipynb'.split('/')
    fs_path = os.path.join(fm.root_dir, *rel_path_list)
    assert path == fs_path

    fm = FileContentsManager(root_dir=str(tmp_path))
    path = fm._get_os_path('test.ipynb')
    fs_path = os.path.join(fm.root_dir, 'test.ipynb')
    assert path == fs_path

    fm = FileContentsManager(root_dir=str(tmp_path))
    path = fm._get_os_path('////test.ipynb')
    fs_path = os.path.join(fm.root_dir, 'test.ipynb')
    assert path == fs_path


def test_checkpoint_subdir(tmp_path):
    subd = 'sub ∂ir'
    cp_name = 'test-cp.ipynb'
    fm = FileContentsManager(root_dir=str(tmp_path))
    tmp_path.joinpath(subd).mkdir()
    cpm = fm.checkpoints
    cp_dir = cpm.checkpoint_path('cp', 'test.ipynb')
    cp_subdir = cpm.checkpoint_path('cp', '/%s/test.ipynb' % subd)
    assert cp_dir != cp_subdir
    assert cp_dir == os.path.join(str(tmp_path), cpm.checkpoint_dir, cp_name)


@pytest.mark.skipif(
    sys.platform == 'win32' and sys.version_info[0] < 3,
    reason="System platform is Windows, version < 3"
)
def test_bad_symlink(tmp_path):
    td = str(tmp_path)

    cm = FileContentsManager(root_dir=td)
    path = 'test bad symlink'
    _make_dir(cm, path)

    file_model = cm.new_untitled(path=path, ext='.txt')

    # create a broken symlink
    symlink(cm, "target", '%s/%s' % (path, 'bad symlink'))
    model = cm.get(path)

    contents = {
        content['name']: content for content in model['content']
    }
    assert 'untitled.txt' in contents
    assert contents['untitled.txt'] == file_model
    assert 'bad symlink' in contents


@pytest.mark.skipif(
    sys.platform == 'win32' and sys.version_info[0] < 3,
    reason="System platform is Windows, version < 3"
)
def test_good_symlink(tmp_path):
    td = str(tmp_path)
    cm = FileContentsManager(root_dir=td)
    parent = 'test good symlink'
    name = 'good symlink'
    path = '{0}/{1}'.format(parent, name)
    _make_dir(cm, parent)

    file_model = cm.new(path=parent + '/zfoo.txt')

    # create a good symlink
    symlink(cm, file_model['path'], path)
    symlink_model = cm.get(path, content=False)
    dir_model = cm.get(parent)
    assert sorted(dir_model['content'], key=lambda x: x['name']) == [symlink_model, file_model]


def test_403(tmp_path):
    if hasattr(os, 'getuid'):
        if os.getuid() == 0:
            raise pytest.skip("Can't test permissions as root")
    if sys.platform.startswith('win'):
        raise pytest.skip("Can't test permissions on Windows")

    td = str(tmp_path)
    cm = FileContentsManager(root_dir=td)
    model = cm.new_untitled(type='file')
    os_path = cm._get_os_path(model['path'])

    os.chmod(os_path, 0o400)
    try:
        with cm.open(os_path, 'w') as f:
            f.write(u"don't care")
    except HTTPError as e:
        assert e.status_code == 403

def test_escape_root(tmp_path):
    td = str(tmp_path)
    cm = FileContentsManager(root_dir=td)
    # make foo, bar next to root
    with open(os.path.join(cm.root_dir, '..', 'foo'), 'w') as f:
        f.write('foo')
    with open(os.path.join(cm.root_dir, '..', 'bar'), 'w') as f:
        f.write('bar')

    with pytest.raises(HTTPError) as e:
        cm.get('..')
    expected_http_error(e, 404)

    with pytest.raises(HTTPError) as e:
        cm.get('foo/../../../bar')
    expected_http_error(e, 404)

    with pytest.raises(HTTPError) as e:
        cm.delete('../foo')
    expected_http_error(e, 404)

    with pytest.raises(HTTPError) as e:
        cm.rename('../foo', '../bar')
    expected_http_error(e, 404)

    with pytest.raises(HTTPError) as e:
        cm.save(model={
            'type': 'file',
            'content': u'',
            'format': 'text',
        }, path='../foo')
    expected_http_error(e, 404)


def test_new_untitled(contents_manager):
    cm = contents_manager
    # Test in root directory
    model = cm.new_untitled(type='notebook')
    assert isinstance(model, dict)
    assert 'name' in model
    assert 'path' in model
    assert 'type' in model
    assert model['type'] == 'notebook'
    assert model['name'] == 'Untitled.ipynb'
    assert model['path'] == 'Untitled.ipynb'

    # Test in sub-directory
    model = cm.new_untitled(type='directory')
    assert isinstance(model, dict)
    assert 'name' in model
    assert 'path' in model
    assert 'type' in model
    assert model['type'] == 'directory'
    assert model['name'] == 'Untitled Folder'
    assert model['path'] == 'Untitled Folder'
    sub_dir = model['path']

    model = cm.new_untitled(path=sub_dir)
    assert isinstance(model, dict)
    assert 'name' in model
    assert 'path' in model
    assert 'type' in model
    assert model['type'] == 'file'
    assert model['name'] == 'untitled'
    assert model['path'] == '%s/untitled' % sub_dir

    # Test with a compound extension
    model = cm.new_untitled(path=sub_dir, ext='.foo.bar')
    assert model['name'] == 'untitled.foo.bar'
    model = cm.new_untitled(path=sub_dir, ext='.foo.bar')
    assert model['name'] == 'untitled1.foo.bar'


def test_modified_date(contents_manager):
    cm = contents_manager
    # Create a new notebook.
    nb, name, path = new_notebook(cm)
    model = cm.get(path)

    # Add a cell and save.
    add_code_cell(model['content'])
    cm.save(model, path)

    # Reload notebook and verify that last_modified incremented.
    saved = cm.get(path)
    assert saved['last_modified'] >= model['last_modified']

    # Move the notebook and verify that last_modified stayed the same.
    # (The frontend fires a warning if last_modified increases on the
    # renamed file.)
    new_path = 'renamed.ipynb'
    cm.rename(path, new_path)
    renamed = cm.get(new_path)
    assert renamed['last_modified'] >= saved['last_modified']


def test_get(contents_manager):
    cm = contents_manager
    # Create a notebook
    model = cm.new_untitled(type='notebook')
    name = model['name']
    path = model['path']

    # Check that we 'get' on the notebook we just created
    model2 = cm.get(path)
    assert isinstance(model2, dict)
    assert 'name' in model2
    assert 'path' in model2
    assert model['name'] == name
    assert model['path'] == path

    nb_as_file = cm.get(path, content=True, type='file')
    assert nb_as_file['path'] == path
    assert nb_as_file['type'] == 'file'
    assert nb_as_file['format'] == 'text'
    assert not isinstance(nb_as_file['content'], dict)

    nb_as_bin_file = cm.get(path, content=True, type='file', format='base64')
    assert nb_as_bin_file['format'] == 'base64'

    # Test in sub-directory
    sub_dir = '/foo/'
    _make_dir(cm, 'foo')
    model = cm.new_untitled(path=sub_dir, ext='.ipynb')
    model2 = cm.get(sub_dir + name)
    assert isinstance(model2, dict)
    assert 'name' in model2
    assert 'path' in model2
    assert 'content' in model2
    assert model2['name'] == 'Untitled.ipynb'
    assert model2['path'] == '{0}/{1}'.format(sub_dir.strip('/'), name)


    # Test with a regular file.
    file_model_path = cm.new_untitled(path=sub_dir, ext='.txt')['path']
    file_model = cm.get(file_model_path)
    expected_model = {
        'content': u'',
        'format': u'text',
        'mimetype': u'text/plain',
        'name': u'untitled.txt',
        'path': u'foo/untitled.txt',
        'type': u'file',
        'writable': True,
    }
    # Assert expected model is in file_model
    for key, value in expected_model.items():
        assert file_model[key] == value
    assert 'created' in file_model
    assert 'last_modified' in file_model

    # Create a sub-sub directory to test getting directory contents with a
    # subdir.
    _make_dir(cm, 'foo/bar')
    dirmodel = cm.get('foo')
    assert dirmodel['type'] == 'directory'
    assert isinstance(dirmodel['content'], list)
    assert len(dirmodel['content']) == 3
    assert dirmodel['path'] == 'foo'
    assert dirmodel['name'] == 'foo'

    # Directory contents should match the contents of each individual entry
    # when requested with content=False.
    model2_no_content = cm.get(sub_dir + name, content=False)
    file_model_no_content = cm.get(u'foo/untitled.txt', content=False)
    sub_sub_dir_no_content = cm.get('foo/bar', content=False)
    assert sub_sub_dir_no_content['path'] == 'foo/bar'
    assert sub_sub_dir_no_content['name'] == 'bar'

    for entry in dirmodel['content']:
        # Order isn't guaranteed by the spec, so this is a hacky way of
        # verifying that all entries are matched.
        if entry['path'] == sub_sub_dir_no_content['path']:
            assert entry == sub_sub_dir_no_content
        elif entry['path'] == model2_no_content['path']:
            assert entry == model2_no_content
        elif entry['path'] == file_model_no_content['path']:
            assert entry == file_model_no_content
        else:
            assert False, "Unexpected directory entry: %s" % entry()

    with pytest.raises(HTTPError):
        cm.get('foo', type='file')


def test_update(contents_manager):
    cm = contents_manager
    # Create a notebook.
    model = cm.new_untitled(type='notebook')
    name = model['name']
    path = model['path']

    # Change the name in the model for rename
    model['path'] = 'test.ipynb'
    model = cm.update(model, path)
    assert isinstance(model, dict)
    assert 'name' in model
    assert 'path' in model
    assert model['name'] == 'test.ipynb'

    # Make sure the old name is gone
    with pytest.raises(HTTPError):
        cm.get(path)

    # Test in sub-directory
    # Create a directory and notebook in that directory
    sub_dir = '/foo/'
    _make_dir(cm, 'foo')
    model = cm.new_untitled(path=sub_dir, type='notebook')
    path = model['path']

    # Change the name in the model for rename
    d = path.rsplit('/', 1)[0]
    new_path = model['path'] = d + '/test_in_sub.ipynb'
    model = cm.update(model, path)
    assert isinstance(model, dict)
    assert 'name' in model
    assert 'path' in model
    assert model['name'] == 'test_in_sub.ipynb'
    assert model['path'] == new_path

    # Make sure the old name is gone
    with pytest.raises(HTTPError):
        cm.get(path)


def test_save(contents_manager):
    cm = contents_manager
    # Create a notebook
    model = cm.new_untitled(type='notebook')
    name = model['name']
    path = model['path']

    # Get the model with 'content'
    full_model = cm.get(path)

    # Save the notebook
    model = cm.save(full_model, path)
    assert isinstance(model, dict)
    assert 'name' in model
    assert 'path' in model
    assert model['name'] == name
    assert model['path'] == path

    # Test in sub-directory
    # Create a directory and notebook in that directory
    sub_dir = '/foo/'
    _make_dir(cm, 'foo')
    model = cm.new_untitled(path=sub_dir, type='notebook')
    name = model['name']
    path = model['path']
    model = cm.get(path)

    # Change the name in the model for rename
    model = cm.save(model, path)
    assert isinstance(model, dict)
    assert 'name' in model
    assert 'path' in model
    assert model['name'] == 'Untitled.ipynb'
    assert model['path'] == 'foo/Untitled.ipynb'


def test_delete(contents_manager):
    cm = contents_manager
    # Create a notebook
    nb, name, path = new_notebook(cm)

    # Delete the notebook
    cm.delete(path)

    # Check that deleting a non-existent path raises an error.
    with pytest.raises(HTTPError):
        cm.delete(path)

    # Check that a 'get' on the deleted notebook raises and error
    with pytest.raises(HTTPError):
        cm.get(path)


def test_rename(contents_manager):
    cm = contents_manager
    # Create a new notebook
    nb, name, path = new_notebook(cm)

    # Rename the notebook
    cm.rename(path, "changed_path")

    # Attempting to get the notebook under the old name raises an error
    with pytest.raises(HTTPError):
        cm.get(path)
    # Fetching the notebook under the new name is successful
    assert isinstance(cm.get("changed_path"), dict)

    # Ported tests on nested directory renaming from pgcontents
    all_dirs = ['foo', 'bar', 'foo/bar', 'foo/bar/foo', 'foo/bar/foo/bar']
    unchanged_dirs = all_dirs[:2]
    changed_dirs = all_dirs[2:]

    for _dir in all_dirs:
        make_populated_dir(cm, _dir)
        check_populated_dir_files(cm, _dir)

    # Renaming to an existing directory should fail
    for src, dest in combinations(all_dirs, 2):
        with pytest.raises(HTTPError) as e:
            cm.rename(src, dest)
        assert expected_http_error(e, 409)

    # Creating a notebook in a non_existant directory should fail
    with pytest.raises(HTTPError) as e:
        cm.new_untitled("foo/bar_diff", ext=".ipynb")
    assert expected_http_error(e, 404)

    cm.rename("foo/bar", "foo/bar_diff")

    # Assert that unchanged directories remain so
    for unchanged in unchanged_dirs:
        check_populated_dir_files(cm, unchanged)

    # Assert changed directories can no longer be accessed under old names
    for changed_dirname in changed_dirs:
        with pytest.raises(HTTPError) as e:
            cm.get(changed_dirname)
        assert expected_http_error(e, 404)
        new_dirname = changed_dirname.replace("foo/bar", "foo/bar_diff", 1)
        check_populated_dir_files(cm, new_dirname)

    # Created a notebook in the renamed directory should work
    cm.new_untitled("foo/bar_diff", ext=".ipynb")


def test_delete_root(contents_manager):
    cm = contents_manager
    with pytest.raises(HTTPError) as e:
        cm.delete('')
    assert expected_http_error(e, 400)


def test_copy(contents_manager):
    cm = contents_manager
    parent = u'å b'
    name = u'nb √.ipynb'
    path = u'{0}/{1}'.format(parent, name)
    _make_dir(cm, parent)

    orig = cm.new(path=path)
    # copy with unspecified name
    copy = cm.copy(path)
    assert copy['name'] == orig['name'].replace('.ipynb', '-Copy1.ipynb')

    # copy with specified name
    copy2 = cm.copy(path, u'å b/copy 2.ipynb')
    assert copy2['name'] == u'copy 2.ipynb'
    assert copy2['path'] == u'å b/copy 2.ipynb'
    # copy with specified path
    copy2 = cm.copy(path, u'/')
    assert copy2['name'] == name
    assert copy2['path'] == name


def test_mark_trusted_cells(contents_manager):
    cm = contents_manager
    nb, name, path = new_notebook(cm)

    cm.mark_trusted_cells(nb, path)
    for cell in nb.cells:
        if cell.cell_type == 'code':
            assert not cell.metadata.trusted

    cm.trust_notebook(path)
    nb = cm.get(path)['content']
    for cell in nb.cells:
        if cell.cell_type == 'code':
            assert cell.metadata.trusted


def test_check_and_sign(contents_manager):
    cm = contents_manager
    nb, name, path = new_notebook(cm)

    cm.mark_trusted_cells(nb, path)
    cm.check_and_sign(nb, path)
    assert not cm.notary.check_signature(nb)

    cm.trust_notebook(path)
    nb = cm.get(path)['content']
    cm.mark_trusted_cells(nb, path)
    cm.check_and_sign(nb, path)
    assert cm.notary.check_signature(nb)



import json


async def test_get_spec(fetch):
    response = await fetch("api", "spec.yaml", method="GET")
    assert response.code == 200


async def test_get_status(fetch):
    response = await fetch("api", "status", method="GET")
    assert response.code == 200
    assert response.headers.get("Content-Type") == "application/json"
    status = json.loads(response.body.decode("utf8"))
    assert sorted(status.keys()) == [
        "connections",
        "kernels",
        "last_activity",
        "started",
    ]
    assert status["connections"] == 0
    assert status["kernels"] == 0
    assert status["last_activity"].endswith("Z")
    assert status["started"].endswith("Z")



import json
import pytest

from jupyter_server.utils import url_path_join


async def test_create_retrieve_config(fetch):
    sample = {'foo': 'bar', 'baz': 73}
    response = await fetch(
        'api', 'config', 'example',
        method='PUT',
        body=json.dumps(sample)
    )
    assert response.code == 204

    response2 = await fetch(
        'api', 'config', 'example',
        method='GET',
    )
    assert response2.code == 200
    assert json.loads(response2.body.decode()) == sample


async def test_modify(fetch):
    sample = {
        'foo': 'bar', 
        'baz': 73,
        'sub': {'a': 6, 'b': 7}, 
        'sub2': {'c': 8}
    }

    modified_sample = {
        'foo': None,  # should delete foo
        'baz': 75,
        'wib': [1,2,3],
        'sub': {'a': 8, 'b': None, 'd': 9},
        'sub2': {'c': None}  # should delete sub2
    }

    diff = {
        'baz': 75, 
        'wib': [1,2,3],
        'sub': {'a': 8, 'd': 9}
    }

    await fetch(
        'api', 'config', 'example',
        method='PUT',
        body=json.dumps(sample)
    )

    response2 = await fetch(
        'api', 'config', 'example',
        method='PATCH',
        body=json.dumps(modified_sample)
    )

    assert response2.code == 200
    assert json.loads(response2.body.decode()) == diff
    

async def test_get_unknown(fetch):
    response = await fetch(
        'api', 'config', 'nonexistant',
        method='GET',
    )
    assert response.code == 200
    assert json.loads(response.body.decode()) == {}


import os, jinja2
from traitlets import Unicode
from jupyter_server.extension.application import ExtensionApp, ExtensionAppJinjaMixin
from .handlers import (DefaultHandler, RedirectHandler,
  ParameterHandler, TemplateHandler, TypescriptHandler, ErrorHandler)

DEFAULT_STATIC_FILES_PATH = os.path.join(os.path.dirname(__file__), "static")
DEFAULT_TEMPLATE_FILES_PATH = os.path.join(os.path.dirname(__file__), "templates")

class SimpleApp1(ExtensionAppJinjaMixin, ExtensionApp):

    # The name of the extension.
    name = "simple_ext1"

    # The url that your extension will serve its homepage.
    extension_url = '/simple_ext1/default'

    # Should your extension expose other server extensions when launched directly?
    load_other_extensions = True

    # Local path to static files directory.
    static_paths = [
        DEFAULT_STATIC_FILES_PATH
    ]

    # Local path to templates directory.
    template_paths = [
        DEFAULT_TEMPLATE_FILES_PATH
    ]

    configA = Unicode('',
        config=True,
        help='Config A example.'
    )

    configB = Unicode('',
        config=True,
        help='Config B example.'
    )

    configC = Unicode('',
        config=True,
        help='Config C example.'
    )

    def initialize_handlers(self):
        self.handlers.extend([
            (r'/{}/default'.format(self.name), DefaultHandler),
            (r'/{}/params/(.+)$'.format(self.name), ParameterHandler),
            (r'/{}/template1/(.*)$'.format(self.name), TemplateHandler),
            (r'/{}/redirect'.format(self.name), RedirectHandler),
            (r'/{}/typescript/?'.format(self.name), TypescriptHandler),
            (r'/{}/(.*)', ErrorHandler)
        ])

    def initialize_settings(self):
        self.log.info('Config {}'.format(self.config))

#-----------------------------------------------------------------------------
# Main entry point
#-----------------------------------------------------------------------------

main = launch_new_instance = SimpleApp1.launch_instance



import re
import pytest
import tornado
from jupyter_server.base.handlers import path_regex


# build regexps that tornado uses:
path_pat = re.compile('^' + '/x%s' % path_regex + '$')

def test_path_regex():
    for path in (
        '/x',
        '/x/',
        '/x/foo',
        '/x/foo.ipynb',
        '/x/foo/bar',
        '/x/foo/bar.txt',
    ):
        assert re.match(path_pat, path)

def test_path_regex_bad():
    for path in (
        '/xfoo',
        '/xfoo/',
        '/xfoo/bar',
        '/xfoo/bar/',
        '/x/foo/bar/',
        '/x//foo',
        '/y',
        '/y/x/foo',
    ):
        assert re.match(path_pat, path) is None


@pytest.mark.parametrize(
    'uri,expected',
    [
        ("/notebooks/mynotebook/", "/notebooks/mynotebook"),
        ("////foo///", "/foo"),
        ("//example.com/", "/example.com"),
        ("/has/param/?hasparam=true", "/has/param?hasparam=true"),
    ]
)
async def test_trailing_slash(uri, expected, http_server_client, auth_header, base_url):
    # http_server_client raises an exception when follow_redirects=False
    with pytest.raises(tornado.httpclient.HTTPClientError) as err:
        await http_server_client.fetch(
            uri,
            headers=auth_header,
            request_timeout=20,
            follow_redirects=False
        )
    # Capture the response from the raised exception value.
    response = err.value.response
    assert response.code == 302
    assert "Location" in response.headers
    assert response.headers["Location"] == expected



import os
import getpass
import pytest
import logging
from unittest.mock import patch

from traitlets import TraitError
from traitlets.tests.utils import check_help_all_output
from jupyter_core.application import NoStart

from jupyter_server.serverapp import (
    ServerApp,
    list_running_servers,
    JupyterPasswordApp,
    JupyterServerStopApp
)
from jupyter_server.auth.security import passwd_check


def test_help_output():
    """jupyter server --help-all works"""
    check_help_all_output('jupyter_server')


def test_server_info_file(tmp_path, configurable_serverapp):
    app = configurable_serverapp(log=logging.getLogger())

    app.write_server_info_file()
    servers = list(list_running_servers(app.runtime_dir))

    assert len(servers) == 1
    sinfo = servers[0]

    assert sinfo['port'] == app.port
    assert sinfo['url'] == app.connection_url
    assert sinfo['version'] == app.version

    app.remove_server_info_file()

    assert list(list_running_servers(app.runtime_dir)) == []
    app.remove_server_info_file


def test_root_dir(tmp_path, configurable_serverapp):
    app = configurable_serverapp(root_dir=str(tmp_path))
    assert app.root_dir == str(tmp_path)


# Build a list of invalid paths
@pytest.fixture(
    params=[
        ('notebooks',),
        ('root', 'dir', 'is', 'missing'),
        ('test.txt',)
    ]
)
def invalid_root_dir(tmp_path, request):
    path = tmp_path.joinpath(*request.param)
    # If the path is a file, create it.
    if os.path.splitext(str(path))[1] != '':
        path.write_text('')
    return str(path)


def test_invalid_root_dir(invalid_root_dir, configurable_serverapp):
    app = configurable_serverapp()
    with pytest.raises(TraitError):
        app.root_dir = invalid_root_dir

@pytest.fixture(
    params=[
        ('/',),
        ('first-level',),
        ('first-level', 'second-level')
    ]
)
def valid_root_dir(tmp_path, request):
    path = tmp_path.joinpath(*request.param)
    if not path.exists():
        # Create path in temporary directory
        path.mkdir(parents=True)
    return str(path)

def test_valid_root_dir(valid_root_dir, configurable_serverapp):
    app = configurable_serverapp(root_dir=valid_root_dir)
    root_dir = valid_root_dir
    # If nested path, the last slash should
    # be stripped by the root_dir trait.
    if root_dir != '/':
        root_dir = valid_root_dir.rstrip('/')
    assert app.root_dir == root_dir


def test_generate_config(tmp_path, configurable_serverapp):
    app = configurable_serverapp(config_dir=str(tmp_path))
    app.initialize(['--generate-config', '--allow-root'])
    with pytest.raises(NoStart):
        app.start()
    assert tmp_path.joinpath('jupyter_server_config.py').exists()


def test_server_password(tmp_path, configurable_serverapp):
    password = 'secret'
    with patch.dict(
        'os.environ', {'JUPYTER_CONFIG_DIR': str(tmp_path)}
        ), patch.object(getpass, 'getpass', return_value=password):
        app = JupyterPasswordApp(log_level=logging.ERROR)
        app.initialize([])
        app.start()
        sv = configurable_serverapp()
        sv.load_config_file()
        assert sv.password != ''
        passwd_check(sv.password, password)


def test_list_running_servers(serverapp, app):
    servers = list(list_running_servers(serverapp.runtime_dir))
    assert len(servers) >= 1




import sys
import time
import json
import shutil
import pytest

import tornado

from nbformat.v4 import new_notebook
from nbformat import writes

from ...utils import expected_http_error

j = lambda r: json.loads(r.body.decode())


@pytest.fixture(params=["MappingKernelManager", "AsyncMappingKernelManager"])
def argv(request):
    if request.param == "AsyncMappingKernelManager" and sys.version_info < (3, 6):
        pytest.skip("Kernel manager is AsyncMappingKernelManager, Python version < 3.6")
    return ["--ServerApp.kernel_manager_class=jupyter_server.services.kernels.kernelmanager." + request.param]


class SessionClient:

    def __init__(self, fetch_callable):
        self.fetch = fetch_callable

    async def _req(self, *args, method, body=None):
        if body is not None:
            body = json.dumps(body)

        r = await self.fetch(
            'api', 'sessions', *args,
            method=method,
            body=body,
            allow_nonstandard_methods=True
        )
        return r

    async def list(self):
        return await self._req(method='GET')

    async def get(self, id):
        return await self._req(id, method='GET')

    async def create(
        self,
        path,
        type='notebook',
        kernel_name='python',
        kernel_id=None):
        body = {
            'path': path,
            'type': type,
            'kernel': {
                'name': kernel_name,
                'id': kernel_id
            }
        }
        return await self._req(method='POST', body=body)

    def create_deprecated(self, path):
        body = {
            'notebook': {
                'path': path
            },
            'kernel': {
                'name': 'python',
                'id': 'foo'
            }
        }
        return self._req(method='POST', body=body)

    def modify_path(self, id, path):
        body = {'path': path}
        return self._req(id, method='PATCH', body=body)

    def modify_path_deprecated(self, id, path):
        body = {'notebook': {'path': path}}
        return self._req(id, method='PATCH', body=body)

    def modify_type(self, id, type):
        body = {'type': type}
        return self._req(id, method='PATCH', body=body)

    def modify_kernel_name(self, id, kernel_name):
        body = {'kernel': {'name': kernel_name}}
        return self._req(id, method='PATCH', body=body)

    def modify_kernel_id(self, id, kernel_id):
        # Also send a dummy name to show that id takes precedence.
        body = {'kernel': {'id': kernel_id, 'name': 'foo'}}
        return self._req(id, method='PATCH', body=body)

    async def delete(self, id):
        return await self._req(id, method='DELETE')

    async def cleanup(self):
        resp = await self.list()
        sessions = j(resp)
        for session in sessions:
            await self.delete(session['id'])
        time.sleep(0.1)



@pytest.fixture
def session_client(root_dir, fetch):
    subdir = root_dir.joinpath('foo')
    subdir.mkdir()

    # Write a notebook to subdir.
    nb = new_notebook()
    nb_str = writes(nb, version=4)
    nbpath = subdir.joinpath('nb1.ipynb')
    nbpath.write_text(nb_str, encoding='utf-8')

    # Yield a session client
    client = SessionClient(fetch)
    yield client

    # Remove subdir
    shutil.rmtree(str(subdir), ignore_errors=True)


def assert_kernel_equality(actual, expected):
    """ Compares kernel models after taking into account that execution_states
        may differ from 'starting' to 'idle'.  The 'actual' argument is the
        current state (which may have an 'idle' status) while the 'expected'
        argument is the previous state (which may have a 'starting' status).
    """
    actual.pop('execution_state', None)
    actual.pop('last_activity', None)
    expected.pop('execution_state', None)
    expected.pop('last_activity', None)
    assert actual == expected


def assert_session_equality(actual, expected):
    """ Compares session models.  `actual` is the most current session,
        while `expected` is the target of the comparison.  This order
        matters when comparing the kernel sub-models.
    """
    assert actual['id'] == expected['id']
    assert actual['path'] == expected['path']
    assert actual['type'] == expected['type']
    assert_kernel_equality(actual['kernel'], expected['kernel'])


async def test_create(session_client):
    # Make sure no sessions exist.
    resp = await session_client.list()
    sessions = j(resp)
    assert len(sessions) == 0

    # Create a session.
    resp = await session_client.create('foo/nb1.ipynb')
    assert resp.code == 201
    new_session = j(resp)
    assert 'id' in new_session
    assert new_session['path'] == 'foo/nb1.ipynb'
    assert new_session['type'] == 'notebook'
    assert resp.headers['Location'] == '/api/sessions/' + new_session['id']

    # Check that the new session appears in list.
    resp = await session_client.list()
    sessions = j(resp)
    assert len(sessions) == 1
    assert_session_equality(sessions[0], new_session)

    # Retrieve that session.
    sid = new_session['id']
    resp = await session_client.get(sid)
    got = j(resp)
    assert_session_equality(got, new_session)

    # Need to find a better solution to this.
    await session_client.cleanup()


async def test_create_file_session(session_client):
    resp = await session_client.create('foo/nb1.py', type='file')
    assert resp.code == 201
    newsession = j(resp)
    assert newsession['path'] == 'foo/nb1.py'
    assert newsession['type'] == 'file'
    await session_client.cleanup()


async def test_create_console_session(session_client):
    resp = await session_client.create('foo/abc123', type='console')
    assert resp.code == 201
    newsession = j(resp)
    assert newsession['path'] == 'foo/abc123'
    assert newsession['type'] == 'console'
    # Need to find a better solution to this.
    await session_client.cleanup()


async def test_create_deprecated(session_client):
    resp = await session_client.create_deprecated('foo/nb1.ipynb')
    assert resp.code == 201
    newsession = j(resp)
    assert newsession['path'] == 'foo/nb1.ipynb'
    assert newsession['type'] == 'notebook'
    assert newsession['notebook']['path'] == 'foo/nb1.ipynb'
    # Need to find a better solution to this.
    await session_client.cleanup()


async def test_create_with_kernel_id(session_client, fetch):
    # create a new kernel
    resp = await fetch('api/kernels', method='POST', allow_nonstandard_methods=True)
    kernel = j(resp)

    resp = await session_client.create('foo/nb1.ipynb', kernel_id=kernel['id'])
    assert resp.code == 201
    new_session = j(resp)
    assert 'id' in new_session
    assert new_session['path'] == 'foo/nb1.ipynb'
    assert new_session['kernel']['id'] == kernel['id']
    assert resp.headers['Location'] == '/api/sessions/{0}'.format(new_session['id'])

    resp = await session_client.list()
    sessions = j(resp)
    assert len(sessions) == 1
    assert_session_equality(sessions[0], new_session)

    # Retrieve it
    sid = new_session['id']
    resp = await session_client.get(sid)
    got = j(resp)
    assert_session_equality(got, new_session)
    # Need to find a better solution to this.
    await session_client.cleanup()

async def test_delete(session_client):
    resp = await session_client.create('foo/nb1.ipynb')
    newsession = j(resp)
    sid = newsession['id']

    resp = await session_client.delete(sid)
    assert resp.code == 204

    resp = await session_client.list()
    sessions = j(resp)
    assert sessions == []

    with pytest.raises(tornado.httpclient.HTTPClientError) as e:
        await session_client.get(sid)
    assert expected_http_error(e, 404)
    # Need to find a better solution to this.
    await session_client.cleanup()

async def test_modify_path(session_client):
    resp = await session_client.create('foo/nb1.ipynb')
    newsession = j(resp)
    sid = newsession['id']

    resp = await session_client.modify_path(sid, 'nb2.ipynb')
    changed = j(resp)
    assert changed['id'] == sid
    assert changed['path'] == 'nb2.ipynb'
    # Need to find a better solution to this.
    await session_client.cleanup()

async def test_modify_path_deprecated(session_client):
    resp = await session_client.create('foo/nb1.ipynb')
    newsession = j(resp)
    sid = newsession['id']

    resp = await session_client.modify_path_deprecated(sid, 'nb2.ipynb')
    changed = j(resp)
    assert changed['id'] == sid
    assert changed['notebook']['path'] == 'nb2.ipynb'
    # Need to find a better solution to this.
    await session_client.cleanup()

async def test_modify_type(session_client):
    resp = await session_client.create('foo/nb1.ipynb')
    newsession = j(resp)
    sid = newsession['id']

    resp = await session_client.modify_type(sid, 'console')
    changed = j(resp)
    assert changed['id'] == sid
    assert changed['type'] == 'console'
    # Need to find a better solution to this.
    await session_client.cleanup()

async def test_modify_kernel_name(session_client, fetch):
    resp = await session_client.create('foo/nb1.ipynb')
    before = j(resp)
    sid = before['id']

    resp = await session_client.modify_kernel_name(sid, before['kernel']['name'])
    after = j(resp)
    assert after['id'] == sid
    assert after['path'] == before['path']
    assert after['type'] == before['type']
    assert after['kernel']['id'] != before['kernel']['id']

    # check kernel list, to be sure previous kernel was cleaned up
    resp = await fetch('api/kernels', method='GET')
    kernel_list = j(resp)
    after['kernel'].pop('last_activity')
    [ k.pop('last_activity') for k in kernel_list ]
    assert kernel_list == [after['kernel']]
    # Need to find a better solution to this.
    await session_client.cleanup()


async def test_modify_kernel_id(session_client, fetch):
    resp = await session_client.create('foo/nb1.ipynb')
    before = j(resp)
    sid = before['id']

    # create a new kernel
    resp = await fetch('api/kernels', method='POST', allow_nonstandard_methods=True)
    kernel = j(resp)

    # Attach our session to the existing kernel
    resp = await session_client.modify_kernel_id(sid, kernel['id'])
    after = j(resp)
    assert after['id'] == sid
    assert after['path'] == before['path']
    assert after['type'] == before['type']
    assert after['kernel']['id'] != before['kernel']['id']
    assert after['kernel']['id'] == kernel['id']

    # check kernel list, to be sure previous kernel was cleaned up
    resp = await fetch('api/kernels', method='GET')
    kernel_list = j(resp)

    kernel.pop('last_activity')
    [ k.pop('last_activity') for k in kernel_list ]
    assert kernel_list == [kernel]

    # Need to find a better solution to this.
    await session_client.cleanup()



import pytest

from traitlets.config import Config
from jupyter_server.services.contents.filecheckpoints import GenericFileCheckpoints


@pytest.fixture
def server_config():
    return {'FileContentsManager': {'checkpoints_class': GenericFileCheckpoints}}


def test_config_did_something(serverapp):
    assert isinstance(serverapp.contents_manager.checkpoints, GenericFileCheckpoints)


import pathlib
from setuptools import setup
from setupbase import (
    get_version, find_packages
)

here = pathlib.Path('.')
version_path = here.joinpath('jupyter_server', '_version.py')
VERSION = get_version(str(version_path))

readme_path = here.joinpath('README.md')
README = readme_path.read_text()

setup_args = dict(
    name             = 'jupyter_server',
    description      = 'The backend—i.e. core services, APIs, and REST endpoints—to Jupyter web applications.',
    long_description = README,
    long_description_content_type='text/markdown',
    version          = VERSION,
    packages         = find_packages('.'),
    include_package_data = True,
    author           = 'Jupyter Development Team',
    author_email     = 'jupyter@googlegroups.com',
    url              = 'http://jupyter.org',
    license          = 'BSD',
    platforms        = "Linux, Mac OS X, Windows",
    keywords         = ['ipython', 'jupyter'],
    classifiers      = [
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
    install_requires = [
        'jinja2',
        'tornado>=5.0',
        'pyzmq>=17',
        'ipython_genutils',
        'traitlets>=4.2.1',
        'jupyter_core>=4.4.0',
        'jupyter_client>=6.1.1',
        'nbformat',
        'nbconvert',
        'Send2Trash',
        'terminado>=0.8.3',
        'prometheus_client',
        "pywin32>=1.0 ; sys_platform == 'win32'"
    ],
    extras_require = {
        'test': ['nose', 'coverage', 'requests', 'nose_warnings_filters',
                 'pytest', 'pytest-cov', 'pytest-tornasync',
                 'pytest-console-scripts', 'ipykernel'],
        'test:sys_platform == "win32"': ['nose-exclude'],
    },
    python_requires = '>=3.6',
    entry_points = {
        'console_scripts': [
            'jupyter-server = jupyter_server.serverapp:main',
        ],
        'pytest11': [
            'pytest_jupyter_server = jupyter_server.pytest_plugin'
        ]
    },
)

if __name__ == '__main__':
    setup(**setup_args)



"""
File-based Checkpoints implementations.
"""
import os
import shutil

from tornado.web import HTTPError

from .checkpoints import (
    Checkpoints,
    GenericCheckpointsMixin,
)
from .fileio import FileManagerMixin

from jupyter_core.utils import ensure_dir_exists
from ipython_genutils.py3compat import getcwd
from traitlets import Unicode

from jupyter_server import _tz as tz


class FileCheckpoints(FileManagerMixin, Checkpoints):
    """
    A Checkpoints that caches checkpoints for files in adjacent
    directories.

    Only works with FileContentsManager.  Use GenericFileCheckpoints if
    you want file-based checkpoints with another ContentsManager.
    """

    checkpoint_dir = Unicode(
        '.ipynb_checkpoints',
        config=True,
        help="""The directory name in which to keep file checkpoints

        This is a path relative to the file's own directory.

        By default, it is .ipynb_checkpoints
        """,
    )

    root_dir = Unicode(config=True)

    def _root_dir_default(self):
        try:
            return self.parent.root_dir
        except AttributeError:
            return getcwd()

    # ContentsManager-dependent checkpoint API
    def create_checkpoint(self, contents_mgr, path):
        """Create a checkpoint."""
        checkpoint_id = u'checkpoint'
        src_path = contents_mgr._get_os_path(path)
        dest_path = self.checkpoint_path(checkpoint_id, path)
        self._copy(src_path, dest_path)
        return self.checkpoint_model(checkpoint_id, dest_path)

    def restore_checkpoint(self, contents_mgr, checkpoint_id, path):
        """Restore a checkpoint."""
        src_path = self.checkpoint_path(checkpoint_id, path)
        dest_path = contents_mgr._get_os_path(path)
        self._copy(src_path, dest_path)

    # ContentsManager-independent checkpoint API
    def rename_checkpoint(self, checkpoint_id, old_path, new_path):
        """Rename a checkpoint from old_path to new_path."""
        old_cp_path = self.checkpoint_path(checkpoint_id, old_path)
        new_cp_path = self.checkpoint_path(checkpoint_id, new_path)
        if os.path.isfile(old_cp_path):
            self.log.debug(
                "Renaming checkpoint %s -> %s",
                old_cp_path,
                new_cp_path,
            )
            with self.perm_to_403():
                shutil.move(old_cp_path, new_cp_path)

    def delete_checkpoint(self, checkpoint_id, path):
        """delete a file's checkpoint"""
        path = path.strip('/')
        cp_path = self.checkpoint_path(checkpoint_id, path)
        if not os.path.isfile(cp_path):
            self.no_such_checkpoint(path, checkpoint_id)

        self.log.debug("unlinking %s", cp_path)
        with self.perm_to_403():
            os.unlink(cp_path)

    def list_checkpoints(self, path):
        """list the checkpoints for a given file

        This contents manager currently only supports one checkpoint per file.
        """
        path = path.strip('/')
        checkpoint_id = "checkpoint"
        os_path = self.checkpoint_path(checkpoint_id, path)
        if not os.path.isfile(os_path):
            return []
        else:
            return [self.checkpoint_model(checkpoint_id, os_path)]

    # Checkpoint-related utilities
    def checkpoint_path(self, checkpoint_id, path):
        """find the path to a checkpoint"""
        path = path.strip('/')
        parent, name = ('/' + path).rsplit('/', 1)
        parent = parent.strip('/')
        basename, ext = os.path.splitext(name)
        filename = u"{name}-{checkpoint_id}{ext}".format(
            name=basename,
            checkpoint_id=checkpoint_id,
            ext=ext,
        )
        os_path = self._get_os_path(path=parent)
        cp_dir = os.path.join(os_path, self.checkpoint_dir)
        with self.perm_to_403():
            ensure_dir_exists(cp_dir)
        cp_path = os.path.join(cp_dir, filename)
        return cp_path

    def checkpoint_model(self, checkpoint_id, os_path):
        """construct the info dict for a given checkpoint"""
        stats = os.stat(os_path)
        last_modified = tz.utcfromtimestamp(stats.st_mtime)
        info = dict(
            id=checkpoint_id,
            last_modified=last_modified,
        )
        return info

    # Error Handling
    def no_such_checkpoint(self, path, checkpoint_id):
        raise HTTPError(
            404,
            u'Checkpoint does not exist: %s@%s' % (path, checkpoint_id)
        )


class GenericFileCheckpoints(GenericCheckpointsMixin, FileCheckpoints):
    """
    Local filesystem Checkpoints that works with any conforming
    ContentsManager.
    """
    def create_file_checkpoint(self, content, format, path):
        """Create a checkpoint from the current content of a file."""
        path = path.strip('/')
        # only the one checkpoint ID:
        checkpoint_id = u"checkpoint"
        os_checkpoint_path = self.checkpoint_path(checkpoint_id, path)
        self.log.debug("creating checkpoint for %s", path)
        with self.perm_to_403():
            self._save_file(os_checkpoint_path, content, format=format)

        # return the checkpoint info
        return self.checkpoint_model(checkpoint_id, os_checkpoint_path)

    def create_notebook_checkpoint(self, nb, path):
        """Create a checkpoint from the current content of a notebook."""
        path = path.strip('/')
        # only the one checkpoint ID:
        checkpoint_id = u"checkpoint"
        os_checkpoint_path = self.checkpoint_path(checkpoint_id, path)
        self.log.debug("creating checkpoint for %s", path)
        with self.perm_to_403():
            self._save_notebook(os_checkpoint_path, nb)

        # return the checkpoint info
        return self.checkpoint_model(checkpoint_id, os_checkpoint_path)

    def get_notebook_checkpoint(self, checkpoint_id, path):
        """Get a checkpoint for a notebook."""
        path = path.strip('/')
        self.log.info("restoring %s from checkpoint %s", path, checkpoint_id)
        os_checkpoint_path = self.checkpoint_path(checkpoint_id, path)

        if not os.path.isfile(os_checkpoint_path):
            self.no_such_checkpoint(path, checkpoint_id)

        return {
            'type': 'notebook',
            'content': self._read_notebook(
                os_checkpoint_path,
                as_version=4,
            ),
        }

    def get_file_checkpoint(self, checkpoint_id, path):
        """Get a checkpoint for a file."""
        path = path.strip('/')
        self.log.info("restoring %s from checkpoint %s", path, checkpoint_id)
        os_checkpoint_path = self.checkpoint_path(checkpoint_id, path)

        if not os.path.isfile(os_checkpoint_path):
            self.no_such_checkpoint(path, checkpoint_id)

        content, format = self._read_file(os_checkpoint_path, format=None)
        return {
            'type': 'file',
            'content': content,
            'format': format,
        }



import os, jinja2
from traitlets import Unicode
from jupyter_server.extension.application import ExtensionApp, ExtensionAppJinjaMixin
from .handlers import ParameterHandler, TemplateHandler, IndexHandler, ErrorHandler

DEFAULT_STATIC_FILES_PATH = os.path.join(os.path.dirname(__file__), "static")
DEFAULT_TEMPLATE_FILES_PATH = os.path.join(os.path.dirname(__file__), "templates")

class SimpleApp2(ExtensionAppJinjaMixin, ExtensionApp):

    # The name of the extension.
    name = "simple_ext2"

    # Te url that your extension will serve its homepage.
    extension_url = '/simple_ext2'

    # Should your extension expose other server extensions when launched directly?
    load_other_extensions = True

    # Local path to static files directory.
    static_paths = [
        DEFAULT_STATIC_FILES_PATH
    ]

    # Local path to templates directory.
    template_paths = [
        DEFAULT_TEMPLATE_FILES_PATH
    ]

    configD = Unicode('',
        config=True,
        help='Config D example.'
    )

    def initialize_handlers(self):
        self.handlers.extend([
            (r'/simple_ext2/params/(.+)$', ParameterHandler),
            (r'/simple_ext2/template', TemplateHandler),
            (r'/simple_ext2/?', IndexHandler),
            (r'/simple_ext2/(.*)', ErrorHandler)
        ])

    def initialize_settings(self):
        self.log.info('Config {}'.format(self.config))

#-----------------------------------------------------------------------------
# Main entry point
#-----------------------------------------------------------------------------

main = launch_new_instance = SimpleApp2.launch_instance
